@@ -1,41 +1,21 @@
 use color_eyre::Result;
 use env_logger::Env;
 use futures::future::try_join_all;
-use log::{error, info};
-use prost::Message;
-use serde::{Deserialize, Serialize};
+use log::info;
 use std::time::Duration;
 use tsyncp::barrier;
 
 const ADDR: &str = "localhost:8000";
 const LEN: usize = 10;
 
-#[derive(
-    Clone, Serialize, Deserialize, rkyv::Serialize, rkyv::Deserialize, rkyv::Archive, Message,
-)]
-struct Dummy {
-    #[prost(string, tag = "1")]
-    field1: String,
-    #[prost(uint64, tag = "2")]
-    field2: u64,
-    #[prost(uint64, tag = "3")]
-    field3: u64,
-}
-
 #[tokio::main]
-async fn main() {
+async fn main() -> Result<()> {
     env_logger::Builder::from_env(Env::default().default_filter_or("info")).init();
 
-    if let Err(error) = try_main().await {
-        error!("Error encountered while running service: \n\n{:?}", error);
-    }
-}
-
-async fn try_main() -> Result<()> {
-    let sender_handles = (0..LEN)
+    let waiter_handles = (0..LEN)
         .map(|_| {
             tokio::spawn(async move {
-                let mut waiter: barrier::Waiter = barrier::wait_to(ADDR)
+                let mut waiter: barrier::Waiter = barrier::waiter_to(ADDR)
                     .retry(Duration::from_millis(500), 100)
                     .await?;
 
@@ -49,9 +29,12 @@ async fn try_main() -> Result<()> {
         })
         .collect::<Vec<_>>();
 
-    let receiver_handle = tokio::spawn(async move {
-        let mut barrier: barrier::Barrier =
-            barrier::block_on(ADDR).limit(LEN).accept_full().await?;
+    let barrier_handle = tokio::spawn(async move {
+        let mut barrier: barrier::Barrier = barrier::barrier_on(ADDR)
+            .limit(LEN)
+            .accept()
+            .to_limit()
+            .await?;
 
         for i in 0..5 {
             tokio::time::sleep(Duration::from_millis(1000)).await;
@@ -64,11 +47,11 @@ async fn try_main() -> Result<()> {
         Ok::<_, color_eyre::Report>(())
     });
 
-    try_join_all(sender_handles)
+    try_join_all(waiter_handles)
         .await?
         .into_iter()
         .collect::<Result<Vec<_>, _>>()?;
-    receiver_handle.await??;
+    barrier_handle.await??;
 
     Ok(())
 }
