@@ -1,9 +1,8 @@
 use color_eyre::Result;
 use env_logger::Env;
 use futures::future::try_join_all;
-use log::{error, info};
+use log::info;
 use prost::Message;
-use serde::{Deserialize, Serialize};
 use std::time::{Duration, Instant};
 use tsyncp::broadcast;
 
@@ -12,9 +11,7 @@ const LEN: usize = 10;
 
 const ADDR: &str = "localhost:8000";
 
-#[derive(
-    Clone, Serialize, Deserialize, rkyv::Serialize, rkyv::Deserialize, rkyv::Archive, Message,
-)]
+#[derive(Message)]
 struct Dummy {
     #[prost(string, tag = "1")]
     field1: String,
@@ -25,33 +22,28 @@ struct Dummy {
 }
 
 #[tokio::main]
-async fn main() {
+async fn main() -> Result<()> {
     env_logger::Builder::from_env(Env::default().default_filter_or("info")).init();
 
-    if let Err(error) = try_main().await {
-        error!("Error encountered while running service: \n\n{:?}", error);
-        // error.downcast_ref::<Box<dyn snafu::ErrorCompat + Send + Sync + std::fmt::Debug>>();
-    }
-}
-
-async fn try_main() -> Result<()> {
-    let receiver_handle = tokio::spawn(async move {
-        let mut receiever: broadcast::ProtobufSender<Dummy> =
-            broadcast::send_on(ADDR).limit(LEN).accept_full().await?;
-
-        // let (mut receiever, _sender) = receiever.split()?;
+    let broadcast_handle = tokio::spawn(async move {
+        let mut tx: broadcast::ProstSender<Dummy> = broadcast::sender_on(ADDR)
+            .limit(LEN)
+            .accept()
+            .to_limit()
+            .await?;
 
         let now = Instant::now();
         for i in 0..COUNT {
-            receiever
-                .broadcast(Dummy {
-                    field1: "hello world".to_string(),
-                    field2: 123213,
-                    field3: i as u64,
-                })
-                .await?;
+            let dummy = Dummy {
+                field1: "hello world".to_string(),
+                field2: 123213,
+                field3: i as u64,
+            };
+
+            tx.send(dummy).await?;
         }
         let duration = Instant::now() - now;
+
         info!("sending {} msgs took {:?}", COUNT, duration);
 
         Ok::<_, color_eyre::Report>(())
@@ -60,15 +52,14 @@ async fn try_main() -> Result<()> {
     let handles = (0..LEN)
         .map(|n| {
             tokio::spawn(async move {
-                let mut sender: broadcast::ProtobufReceiver<Dummy> = broadcast::recv_to(ADDR)
+                let mut rx: broadcast::ProstReceiver<Dummy> = broadcast::receiver_to(ADDR)
                     .retry(Duration::from_millis(500), 100)
                     .await?;
-                info!("{n}: addr: {}", sender.local_addr());
 
                 let mut i = 0;
 
                 let now = Instant::now();
-                while let Some(Ok(_)) = sender.recv().await {
+                while let Some(Ok(_)) = rx.recv().await {
                     i += 1;
 
                     if i % (COUNT / 10) == 0 {
@@ -87,7 +78,7 @@ async fn try_main() -> Result<()> {
         })
         .collect::<Vec<_>>();
 
-    receiver_handle.await??;
+    broadcast_handle.await??;
     try_join_all(handles)
         .await?
         .into_iter()
