@@ -1,3 +1,11 @@
+//! Contains [BuilderFuture] which builds [channel::Channel](super::Channel) when `.await`ed.
+//!
+//! [BuilderFuture] is returned by [channel_on](super::channel_on) or [channel_to](super::channel_to) function.
+//!
+//! Before awaiting the future, you can chain other methods on it to configure the Channel.
+//!
+//! To see all available configurations, see [BuilderFuture].
+
 use super::Channel;
 use crate::util::{frame_codec::VariedLengthDelimitedCodec, Framed};
 use errors::*;
@@ -12,12 +20,12 @@ use std::time::Duration;
 use tokio::net::{TcpSocket, TcpStream};
 use tokio::task::JoinError;
 
-pub type Result<T, E = ChannelBuilderError> = std::result::Result<T, E>;
+pub(crate) type Result<T, E = BuilderError> = std::result::Result<T, E>;
 
 pub(crate) fn new<A: 'static + Clone + Send + ToSocketAddrs, T, E>(
     addr: A,
     listening: bool,
-) -> ChannelBuilderFuture<
+) -> BuilderFuture<
     A,
     T,
     E,
@@ -29,7 +37,7 @@ pub(crate) fn new<A: 'static + Clone + Send + ToSocketAddrs, T, E>(
     let filter = |_| true;
     let tcp_settings = TcpSettings::default();
 
-    ChannelBuilderFuture {
+    BuilderFuture {
         addr: addr.clone(),
         listening,
         max_retries,
@@ -48,9 +56,38 @@ pub(crate) fn new<A: 'static + Clone + Send + ToSocketAddrs, T, E>(
     }
 }
 
+/// Future returned by [channel_on(_)](crate::channel::channel_on) and [channel_to(_)](crate::channel::channel_to)
+/// to configure and build [Channel](super::Channel).
+///
+/// You can chain any number of configurations to the future:
+///
+/// ```no_run
+/// use tsyncp::channel;
+/// use serde::{Serialize, Deserialize};
+/// use std::time::Duration;
+///
+/// #[derive(Debug, Serialize, Deserialize)]
+/// struct Dummy;
+///
+/// #[tokio::main]
+/// async fn main() -> color_eyre::Result<()> {
+///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_on("localhost:8000")
+///         .retry(Duration::from_millis(500), 100)
+///         .set_tcp_linger(Some(Duration::from_millis(10_000)))
+///         .set_tcp_ttl(60_000)
+///         .set_tcp_nodelay(true)
+///         .set_tcp_reuseaddr(true)
+///         .set_tcp_reuseport(true)
+///         .set_tcp_send_buffer_size(8 * 1024 * 1024)
+///         .set_tcp_recv_buffer_size(8 * 1024 * 1024)
+///         .await?;
+///
+///     Ok(())
+/// }
+/// ```
 #[derive(Debug)]
 #[pin_project]
-pub struct ChannelBuilderFuture<A, T, E, Filter, Fut, S = TcpStream> {
+pub struct BuilderFuture<A, T, E, Filter, Fut, S = TcpStream> {
     addr: A,
     listening: bool,
     max_retries: Option<usize>,
@@ -73,27 +110,70 @@ struct TcpSettings {
     ttl: Option<u32>,
 }
 
-impl<A, T, E, Filter, Fut> ChannelBuilderFuture<A, T, E, Filter, Fut>
+impl<A, T, E, Filter, Fut> BuilderFuture<A, T, E, Filter, Fut>
 where
     A: 'static + Clone + Send + ToSocketAddrs,
     Filter: Clone + Fn(SocketAddr) -> bool,
 {
+    /// Retry connecting to remote address for `max_retries` with the interval
+    /// `retry_sleep_duration`.
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    /// use std::time::Duration;
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_to("localhost:8000")
+    ///         .retry(Duration::from_millis(500), 100)
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     pub fn retry(
         mut self,
         retry_sleep_duration: Duration,
         max_retries: usize,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
         self.max_retries.replace(max_retries);
         self.retry_sleep_duration = retry_sleep_duration;
 
         self.refresh()
     }
 
+    /// When accepting a connection, filter by the given closure.
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    /// use std::time::Duration;
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_on("localhost:8000")
+    ///         .filter(|a| a.port() % 2 == 0)
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     pub fn filter<Filter2: Clone + Fn(SocketAddr) -> bool>(
         self,
         filter: Filter2,
-    ) -> ChannelBuilderFuture<A, T, E, Filter2, impl Future<Output = Result<Channel<T, E>>>> {
-        let ChannelBuilderFuture {
+    ) -> BuilderFuture<A, T, E, Filter2, impl Future<Output = Result<Channel<T, E>>>> {
+        let BuilderFuture {
             addr,
             listening,
             max_retries,
@@ -102,7 +182,7 @@ where
             ..
         } = self;
 
-        ChannelBuilderFuture {
+        BuilderFuture {
             addr: addr.clone(),
             listening,
             max_retries,
@@ -121,15 +201,57 @@ where
         }
     }
 
+    /// Set tcp reuseaddr for the connection made on this channel.
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_to("localhost:8000")
+    ///         .set_tcp_reuseaddr(true)
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     pub fn set_tcp_reuseaddr(
         mut self,
         reuseaddr: bool,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
         self.tcp_settings.reuseaddr.replace(reuseaddr);
 
         self.refresh()
     }
 
+    /// Set tcp reuseport for the connection made on this channel.
+    ///
+    /// *Warning:* only available to unix targets excluding "solaris" and "illumos".
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_to("localhost:8000")
+    ///         .set_tcp_reuseport(true)
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     #[cfg(all(unix, not(target_os = "solaris"), not(target_os = "illumos")))]
     #[cfg_attr(
         docsrs,
@@ -138,52 +260,153 @@ where
     pub fn set_tcp_reuseport(
         mut self,
         reuseport: bool,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
         self.tcp_settings.reuseport.replace(reuseport);
 
         self.refresh()
     }
 
+    /// Set tcp linger for the connection made on this channel.
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    /// use std::time::Duration;
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_to("localhost:8000")
+    ///         .set_tcp_linger(Some(Duration::from_millis(10_000)))
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     pub fn set_tcp_linger(
         mut self,
         dur: Option<Duration>,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
         self.tcp_settings.linger.replace(dur);
 
         self.refresh()
     }
 
+    /// Set tcp nodelay for the connection made on this channel.
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_on("localhost:8000")
+    ///         .set_tcp_nodelay(true)
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     pub fn set_tcp_nodelay(
         mut self,
         nodelay: bool,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
         self.tcp_settings.nodelay.replace(nodelay);
 
         self.refresh()
     }
 
+    /// Set tcp ttl for the connection made on this channel.
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_to("localhost:8000")
+    ///         .set_tcp_ttl(60_000)
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     pub fn set_tcp_ttl(
         mut self,
         ttl: u32,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
         self.tcp_settings.ttl.replace(ttl);
 
         self.refresh()
     }
 
+    /// Set tcp recv_buffer_size for the connection made on this channel.
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_on("localhost:8000")
+    ///         .set_tcp_recv_buffer_size(8 * 1024 * 1024)
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     pub fn set_tcp_recv_buffer_size(
         mut self,
         size: u32,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
         self.tcp_settings.recv_buffer_size.replace(size);
 
         self.refresh()
     }
 
+    /// Set tcp send_buffer_size for the connection made on this channel.
+    ///
+    /// ### Example:
+    ///
+    /// ```no_run
+    /// use tsyncp::channel;
+    /// use serde::{Serialize, Deserialize};
+    ///
+    /// #[derive(Debug, Serialize, Deserialize)]
+    /// struct Dummy;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> color_eyre::Result<()> {
+    ///     let mut ch: channel::JsonChannel<Dummy> = channel::channel_on("localhost:8000")
+    ///         .set_tcp_send_buffer_size(8 * 1024 * 1024)
+    ///         .await?;
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
     pub fn set_tcp_send_buffer_size(
         mut self,
         size: u32,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
         self.tcp_settings.send_buffer_size.replace(size);
 
         self.refresh()
@@ -191,8 +414,8 @@ where
 
     fn refresh(
         self,
-    ) -> ChannelBuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
-        let ChannelBuilderFuture {
+    ) -> BuilderFuture<A, T, E, Filter, impl Future<Output = Result<Channel<T, E>>>> {
+        let BuilderFuture {
             addr,
             listening,
             max_retries,
@@ -202,7 +425,7 @@ where
             ..
         } = self;
 
-        ChannelBuilderFuture {
+        BuilderFuture {
             addr: addr.clone(),
             listening,
             max_retries,
@@ -223,7 +446,7 @@ where
 }
 
 impl<A, T, E, Filter, Fut: Future<Output = Result<Channel<T, E, S>>>, S> Future
-    for ChannelBuilderFuture<A, T, E, Filter, Fut, S>
+    for BuilderFuture<A, T, E, Filter, Fut, S>
 {
     type Output = Result<Channel<T, E, S>>;
 
@@ -366,82 +589,83 @@ fn get_socket(addr: &SocketAddr, tcp_settings: &TcpSettings) -> Result<TcpSocket
     Ok(socket)
 }
 
+#[allow(missing_docs)]
 pub mod errors {
     use super::*;
     use snafu::Snafu;
 
     #[derive(Debug, Snafu)]
     #[snafu(visibility(pub(super)))]
-    pub enum ChannelBuilderError {
-        #[snafu(display("[ChannelBuilderError] Encountered unexpected error on spawned task for converting to socket addr"))]
+    pub enum BuilderError {
+        #[snafu(display("[BuilderError] Encountered unexpected error on spawned task for converting to socket addr"))]
         SpawnJoin {
             /// source Error
             source: JoinError,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to convert input to socket address"))]
+        #[snafu(display("[BuilderError] Failed to convert input to socket address"))]
         ToSocketAddrs {
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to create Tcp Socket for {addr}"))]
+        #[snafu(display("[BuilderError] Failed to create Tcp Socket for {addr}"))]
         NewSocketError {
             addr: SocketAddr,
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to set reuseaddr for {addr}"))]
+        #[snafu(display("[BuilderError] Failed to set reuseaddr for {addr}"))]
         SetReuseAddr {
             addr: SocketAddr,
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to set reuseport for {addr}"))]
+        #[snafu(display("[BuilderError] Failed to set reuseport for {addr}"))]
         SetReusePort {
             addr: SocketAddr,
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to set linger for {addr}"))]
+        #[snafu(display("[BuilderError] Failed to set linger for {addr}"))]
         SetLinger {
             addr: SocketAddr,
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to set recv_buffer_size for {addr}"))]
+        #[snafu(display("[BuilderError] Failed to set recv_buffer_size for {addr}"))]
         SetRecvBufferSize {
             addr: SocketAddr,
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to set send_buffer_size for {addr}"))]
+        #[snafu(display("[BuilderError] Failed to set send_buffer_size for {addr}"))]
         SetSendBufferSize {
             addr: SocketAddr,
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to set nodelay for {addr}"))]
+        #[snafu(display("[BuilderError] Failed to set nodelay for {addr}"))]
         SetNodelay {
             addr: SocketAddr,
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to set ttl for {addr}"))]
+        #[snafu(display("[BuilderError] Failed to set ttl for {addr}"))]
         SetTtl {
             addr: SocketAddr,
             /// source IO Error
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to connect to {addr}"))]
+        #[snafu(display("[BuilderError] Failed to connect to {addr}"))]
         Connecting {
             addr: SocketAddr,
             /// source IO Error
@@ -449,7 +673,7 @@ pub mod errors {
             backtrace: Backtrace,
         },
         /// returned from invalid inner IO Error
-        #[snafu(display("[ChannelBuilderError] Failed to bind on {addr}"))]
+        #[snafu(display("[BuilderError] Failed to bind on {addr}"))]
         Binding {
             addr: SocketAddr,
             /// source IO Error
@@ -457,7 +681,7 @@ pub mod errors {
             backtrace: Backtrace,
         },
         /// returned from invalid inner IO Error
-        #[snafu(display("[ChannelBuilderError] Failed to listen on {addr}"))]
+        #[snafu(display("[BuilderError] Failed to listen on {addr}"))]
         Listening {
             addr: SocketAddr,
             /// source IO Error
@@ -466,7 +690,7 @@ pub mod errors {
         },
         /// returned from invalid inner IO Error
         #[snafu(display(
-            "[ChannelBuilderError] Encountered IO Error while accepting connections on {addr}"
+            "[BuilderError] Encountered IO Error while accepting connections on {addr}"
         ))]
         Accepting {
             addr: SocketAddr,
@@ -474,7 +698,7 @@ pub mod errors {
             source: io::Error,
             backtrace: Backtrace,
         },
-        #[snafu(display("[ChannelBuilderError] Failed to get local addr for listener on {addr}"))]
+        #[snafu(display("[BuilderError] Failed to get local addr for listener on {addr}"))]
         LocalAddr {
             addr: SocketAddr,
             source: io::Error,
