@@ -1,13 +1,12 @@
 use bytes::{Buf, BufMut, Bytes, BytesMut};
 use errors::*;
-use std::io;
 use tokio::io::{AsyncRead, AsyncWrite};
 use tokio_util::codec::{Decoder, Encoder, Framed};
 
 /// default MAX_FRAME_LENGTH for codec
 pub const MAX_FRAME_LENGTH: usize = 8 * 1024 * 1024;
 
-type Result<T, E = LengthDelimitedCodecError> = std::result::Result<T, E>;
+type Result<T, E = CodecError> = std::result::Result<T, E>;
 
 /// Length delimited bytes-to-frame encoding/decoding scheme that uses first n bits to represent header bytes' length. (1 <= n <= 3)
 ///
@@ -130,6 +129,8 @@ impl VariedLengthDelimitedCodec {
     }
 
     /// Returns `Self` with custom `max_frame_length`
+    ///
+    /// Defaults to `MAX_FRAME_LENGTH = 8 * 1024 * 1024`
     pub fn with_max_frame_length(max_frame_length: usize) -> Self {
         Self {
             max_frame_length,
@@ -139,7 +140,7 @@ impl VariedLengthDelimitedCodec {
 
     /// Create a configured length delimited `Framed`
     ///
-    /// # Examples
+    /// # Example
     ///
     /// ```
     /// use tokio::io::{AsyncRead, AsyncWrite};
@@ -186,10 +187,11 @@ impl VariedLengthDelimitedCodec {
         let payload_len = src.get_uint(header_len) as usize;
 
         if payload_len > self.max_frame_length {
-            return Err(LengthDelimitedCodecError::InvalidDecodingFrameLength {
+            return InvalidDecodingFrameLengthSnafu {
                 len: payload_len,
                 max_frame_length: self.max_frame_length,
-            });
+            }
+            .fail();
         }
 
         // Ensure that the buffer has enough space to read the incoming payload
@@ -211,7 +213,7 @@ impl VariedLengthDelimitedCodec {
 
 impl Decoder for VariedLengthDelimitedCodec {
     type Item = BytesMut;
-    type Error = LengthDelimitedCodecError;
+    type Error = CodecError;
 
     fn decode(&mut self, src: &mut BytesMut) -> Result<Option<Self::Item>> {
         let payload_len = match self.state {
@@ -241,16 +243,17 @@ impl Decoder for VariedLengthDelimitedCodec {
 }
 
 impl Encoder<Bytes> for VariedLengthDelimitedCodec {
-    type Error = LengthDelimitedCodecError;
+    type Error = CodecError;
 
     fn encode(&mut self, data: Bytes, dst: &mut BytesMut) -> Result<()> {
         let payload_len = data.len();
 
         if payload_len > self.max_frame_length {
-            return Err(LengthDelimitedCodecError::InvalidEncodingFrameLength {
+            return InvalidEncodingFrameLengthSnafu {
                 len: payload_len,
                 max_frame_length: self.max_frame_length,
-            });
+            }
+            .fail();
         }
 
         // single byte max length
@@ -275,15 +278,15 @@ impl Encoder<Bytes> for VariedLengthDelimitedCodec {
     }
 }
 
+#[allow(missing_docs)]
 pub mod errors {
-    use super::*;
-    use snafu::Snafu;
+    use snafu::{Backtrace, GenerateImplicitData, Snafu};
+    use std::io;
 
     /// Codec's error type
-    /// TODO: separate error types to `EncodeError` and `DecodeError`
     #[derive(Debug, Snafu)]
     #[snafu(visibility(pub(super)))]
-    pub enum LengthDelimitedCodecError {
+    pub enum CodecError {
         /// Invalid length in frame header was received while decoding frame.
         #[snafu(display("Received invalid frame length {len} while decoding; bytes' length must be greater 0 and less than {max_frame_length}"))]
         InvalidDecodingFrameLength {
@@ -291,6 +294,7 @@ pub mod errors {
             len: usize,
             /// max frame length
             max_frame_length: usize,
+            backtrace: Backtrace,
         },
         /// Invalid length in frame header was received while encoding frame.
         #[snafu(display("Received invalid frame length {len} while encoding; bytes' length must be greater 0 and less than {max_frame_length}"))]
@@ -299,29 +303,123 @@ pub mod errors {
             len: usize,
             /// max frame length
             max_frame_length: usize,
+            backtrace: Backtrace,
         },
-        /// returned from invalid inner IO Error
+        /// Returned from invalid inner IO Error.
         #[snafu(display("Encountered IO Error while decoding frame"))]
         IoError {
-            /// source IO Error
+            /// Source IO Error
             source: io::Error,
+            backtrace: Backtrace,
         },
     }
 
-    impl From<io::Error> for LengthDelimitedCodecError {
+    impl From<io::Error> for CodecError {
         fn from(src: io::Error) -> Self {
-            Self::IoError { source: src }
+            Self::IoError {
+                source: src,
+                backtrace: Backtrace::generate(),
+            }
         }
     }
 
-    impl LengthDelimitedCodecError {
+    impl CodecError {
+        /// Try to view the error as [std::io::Error].
+        ///
+        /// This is useful to see if the returned error is from the underlying TCP connection.
+        /// This method will be bubbled up with the error, and also be available at the highest
+        /// level.
         pub fn as_io(&self) -> Option<&io::Error> {
-            if let Self::IoError { source } = self {
+            if let Self::IoError { source, .. } = self {
                 return Some(source);
             }
 
             None
         }
+
+        /// Try converting the error as [std::io::Error].
+        ///
+        /// This is useful to see if the returned error is from the underlying TCP connection.
+        /// This method will be bubbled up with the error, and also be available at the highest
+        /// level.
+        pub fn into_io(self) -> Option<io::Error> {
+            if let Self::IoError { source, .. } = self {
+                return Some(source);
+            }
+
+            None
+        }
+
+        /// Check if the error is a connection error.
+        ///
+        /// Returns `true` if the error either `reset`, `refused`, `aborted`, `not connected`, or
+        /// `broken pipe`.
+        ///
+        /// This is useful to see if the returned error is from the underlying TCP connection.
+        /// This method will be bubbled up with the error, and also be available at the highest
+        /// level.
+        pub fn is_connection_error(&self) -> bool {
+            self.is_connection_reset()
+                || self.is_connection_refused()
+                || self.is_connection_aborted()
+                || self.is_not_connected()
+                || self.is_broken_pipe()
+        }
+
+        /// Check if the error is from connection reset.
+        ///
+        /// This is useful to see if the returned error is from the underlying TCP connection.
+        /// This method will be bubbled up with the error, and also be available at the highest
+        /// level.
+        pub fn is_connection_reset(&self) -> bool {
+            self.as_io()
+                .map(|e| e.kind() == io::ErrorKind::ConnectionReset)
+                .unwrap_or_default()
+        }
+
+        /// Check if the error is from connection refused.
+        ///
+        /// This is useful to see if the returned error is from the underlying TCP connection.
+        /// This method will be bubbled up with the error, and also be available at the highest
+        /// level.
+        pub fn is_connection_refused(&self) -> bool {
+            self.as_io()
+                .map(|e| e.kind() == io::ErrorKind::ConnectionRefused)
+                .unwrap_or_default()
+        }
+
+        /// Check if the error is from connection aborted.
+        ///
+        /// This is useful to see if the returned error is from the underlying TCP connection.
+        /// This method will be bubbled up with the error, and also be available at the highest
+        /// level.
+        pub fn is_connection_aborted(&self) -> bool {
+            self.as_io()
+                .map(|e| e.kind() == io::ErrorKind::ConnectionAborted)
+                .unwrap_or_default()
+        }
+
+        /// Check if the error is from not connected.
+        ///
+        /// This is useful to see if the returned error is from the underlying TCP connection.
+        /// This method will be bubbled up with the error, and also be available at the highest
+        /// level.
+        pub fn is_not_connected(&self) -> bool {
+            self.as_io()
+                .map(|e| e.kind() == io::ErrorKind::NotConnected)
+                .unwrap_or_default()
+        }
+
+        /// Check if the error is from broken pipe.
+        ///
+        /// This is useful to see if the returned error is from the underlying TCP connection.
+        /// This method will be bubbled up with the error, and also be available at the highest
+        /// level.
+        pub fn is_broken_pipe(&self) -> bool {
+            self.as_io()
+                .map(|e| e.kind() == io::ErrorKind::BrokenPipe)
+                .unwrap_or_default()
+        }
     }
 }
 
@@ -335,7 +433,7 @@ mod tests {
     use serde::{Deserialize, Serialize};
     use std::iter;
 
-    #[derive(Debug, Clone, Serialize, Deserialize, Dummy, PartialEq)]
+    #[derive(Debug, Serialize, Deserialize, Dummy, PartialEq)]
     struct ExampleStruct {
         #[dummy(faker = "1..2000")]
         example_field1: usize,
